#!/bin/bash -u

# the script creates hough images for science frames.
# 
# First version : 31/01/2007
# 

# 09.10.2007
# Each call of this script creates its own version of the
# initial HOUGH weight image. This avoids potential problems
# during parallel processing.
#
# 16.09.2008
# Temporary files are cleaned at the end of the script
#
# 07.12.2008:
# I cleaned up the script a bit
#
# 10.10.2009:
# the script now accepts a file with exposures to process.
# This allows us to split satellite track detection in junks
# of images. This was necessary because this step creates very
# large files and hence often disk space problems for larger
# runs.
#
# 09.01.2013:
# I made the script more robust to non-existing files
#
# 24.09.2014:
# I made file inclusions (sourcing) more robust to make sure that
# files are included from the present work directory.

# $1: main directory
# $2: science dir.
# $3: image extension (ext) on ..._iext.fits (i is the chip number)
# $4: file with images to process (complete path; OPTIONAL argument)
#     on each line the file contains the prefix of one multi-chip
#     file. If filenames are like 865432p_1C.fits, the prefix
#     would be 865432p
# $#: chips to be processed

. ./${INSTRUMENT:?}.ini
. ./bash_functions.include
theli_start "$*" "${!#}"

# hard-coded parameters 
ELL=0.3
WEIGHTDIR="/$1/WEIGHTS"

# export PWD for progs.ini to be found within perl scripts
PROGS_DIR=${PWD}
export PROGS_DIR;

cd /$1/$2

for CHIP in ${!#}
do
  test -f ${TEMPDIR}/hough_images_$$ && rm ${TEMPDIR}/hough_images_$$

  if [ $# -eq 5 ]; then
    while read IMAGE
    do
      test -f ${IMAGE}_${CHIP}$3.fits && \
        echo "${IMAGE}_${CHIP}$3.fits" >> ${TEMPDIR}/hough_images_$$
    done < $4  
  else
    ${P_FIND} /$1/$2/ -name \*_${CHIP}$3.fits > ${TEMPDIR}/hough_images_$$
  fi

  # do we need to do something at all?
  if [ -s ${TEMPDIR}/hough_images_$$ ]; then
    while read FILE
    do
      # Create the segmentation image
      BASE=`basename ${FILE} $3.fits`
      WEIGHT=${WEIGHTDIR}/${BASE}$3.weight.fits
      SEG=${BASE}$3_SEG.fits
      
      if [ ! -f ${SEG} ]; then    
        cd /$1/$2
        perl ${S_STRACK_SEG} \
             -image ${FILE} \
             -weight ${WEIGHT} \
             -ell ${ELL} \
             -seeing \
             -instrum ${INSTRUMENT} 
      fi
          
      HSN=${BASE}$3_SEG_hSN.fits
  
      # create the hough image
      if [ ! -f ${HSN} ]; then
        perl ${S_STRACK_HOUGH} \
             -seg ${SEG} \
             -instrum ${INSTRUMENT} \
             -hweight \
             -check 0.8 \
             -hweight_fin HTweight_${INSTRUMENT}.fits_$$
      fi
    done < ${TEMPDIR}/hough_images_$$
  else
    theli_warn "No images to process for Chip ${CHIP}"  
  fi
done

# clean up and bye:
test -f ${TEMPDIR}/hough_images_$$ && rm ${TEMPDIR}/hough_images_$$

cd ${PROGS_DIR}

theli_end "${!#}"
exit 0;
