#!/bin/bash -u

# the script create_strack_mask_exposurepara.sh creates hough images
# for science frames.  It is a single-machine, parallel version from
# Patricks script create_strack_mask.sh Unlike most other parallel
# THELI scripts it splits the job by images and NOT by chips!
#
# HISTORY INFORMAION:
# 
# First version : 09/09/2008
# 
# 10.10.2009:
# the script now accepts a file with exposures to process.
# This allows us to split satellite track detection in junks
# of images. This was necessary because this step creates very
# large files and hence often disk space problems for larger
# runs.
#
# 09.01.2013:
# I made the script more robust to non-existing files
#
# 17.09.2013:
# I set the minimum width of tracks to 30 pixels. We saw with
# KIDS data that the default of 12 pixlels was not enough.
#
# 24.09.2014:
# I made file inclusions (sourcing) more robust to make sure that
# files are included from the present work directory.

# TODOS:
# - Remove implicit assumption on the WEIGHTS directory
# - Remove implicit assumption on the presence of the
#   first chip

# $1: main directory
# $2: science dir.
# $3: image extension (ext) on ..._iext.fits (i is the chip number)
# $4: file with images to process (complete path; OPTIONAL argument)
#     on each line the file contains the prefix of one multi-chip
#     file. If filenames are like 865432p_1C.fits, the prefix
#     would be 865432p

. ./${INSTRUMENT:?}.ini
. ./bash_functions.include
theli_start "$*"

if [ $# -lt 3 ] && [ $# -gt 4 ]; then
  theli_error "Wrong number of command line arguments! Exiting!"
  exit 1;
fi

PROGS_DIR=$PWD
export PROGS_DIR;
export PIPESOFT; # defined in progs.ini which is included by
                 # ${INSTRUMENT}.ini
WEIGHTDIR="/$1/WEIGHTS"

cd /$1/$2

# split the exposures to available processors:
i=0
while [ ${i} -lt ${NPARA} ]
do
  if [ $# -eq 4 ]; then
    FILES[$i]=`sort $4 | ${P_GAWK} '(NR%'${NPARA}' == '$i')'` 
  else
    # the 'sort' in the following pipeline ensures that the order of files is
    # always the same. This is not ensured by the 'find' command!
    #
    # NOTE that we assume that chip Nr. 1 always exists here
    # (THIS SHOULD BE MADE MORE GENERAL)
    FILES[$i]=`${P_FIND} . -maxdepth 1 -name \*_1${3}.fits  |\
               sort | ${P_GAWK} '(NR%'${NPARA}' == '$i')'` 
  fi
  NFILES[$i]=`echo ${FILES[$i]} | wc -w` 
  j=$(( ${i} + 1 ))

  if [ ${NFILES[$i]} -gt 0 ]; then
    echo -e "Starting Job $j. It has ${NFILES[$i]} files to process!\n"
   
    # Loop on the exposures 
    {
      k=1
      for file in ${FILES[$i]}
      do
        # the following basename command does nothing in case
        # that files are given with a prefix a file (optional
        # fourth command line argument)
        BASE=`basename ${file} _1$3.fits`
        echo "Job ${j} (${k}/${NFILES[$i]}): working on image ${BASE}"
  
        perl ${S_STRACK_HOUGH_DETECT} \
             -root ${BASE} -ext ${3} -min_width 30\
             -instrum ${INSTRUMENT} -noim -dthresh 0.8 -norefine
  
        k=$(( $k + 1 ))
      done
    } &
  else
    theli_warn "No images to process for Job $j"
  fi
  i=$(( $i + 1 ))
done

# only continue if all launched background jobs are
# finished:
wait;

# copy the reg files at the right place
test -d ./STRACK || mkdir ./STRACK
mv ./*.reg ./STRACK

# and bye
cd ${PROGS_DIR}  
theli_end
exit 0;
