#!/usr/bin/env perl -w

############################################################################
# Detect the tracks on the hough images + check them
#                                       + build the flag images
# input : root : root name of the image
#       : ext  : extention to use for stracks detection (def = OFCS)
#       : ext1 : extention to use for stracks detection (def = _hSN1.fits)
#       : mthresh :  threshold for masking (def = 0.8)
#       : small :  threshold for small track check (def = 0.8)
#       : radius : radius to grow the flag image (def = 0)
#       : ext_out : output extention (def = _flag.fits)
#       : boarder : check image boarder
#       : log     : log file
#       : noclean : do not clean temporary files (debug)
#       : instrum : instrumentation
#       : norefine: don't try to refine the position of the line
#       : nocheck : don't check before adding missing tracks
############################################################################

# HISTORY INFORMATION (changes by Thomas Erben from Patricks original code):
#
# 09.09.2008:
# - I cleaned the code a bit (removal of old, commented section)
#
# - All temporary files get now uniq filenames (including procedd ID)
#   This allows us to call several versions of this script in the same
#   directory
#
# 16.09.2008:
# - I introduced a work-around for occuring NaN values in the width of tracks
#   and subsequent program crashes when they should be added to region files.
#
# 22.08.2013:
# I increased the grouping check radius from 50 to 60 pixels. It turned out
# for KIDS that 50 is a bit too small for several tracks.
#
# 16.09.2013:
# I fixed a major bug in the treatment of 'undetected' tracks. My test
# for 'NaN' values always returned 'True'! This is probably one reason
# for a lot of unmasked short tracks!
#
# 25.07.2014:
# function check_vertical_layout:
# I updated the function to also treat the DECam@CTIO camera. That
# camera has a very irregular chip pattern and horizontal walks from
# a chip do not lead to a uniqe neigbouring chip. The function had
# to be updated to deal with this correctly.

# TODO:
# Cleanly treat occuring NaN values!

if (defined $ENV{'PERL_PACK'}) {
    use lib $ENV{'PERL_PACK'};
}

$homedir = $ENV{'HOME'};
push(@INC, $ENV{"PIPESOFT"} . "/stracktools/scripts");

require "sex_utils.pl";
require "CFH12K_lib_port.pl";
require "progs_lib.pl";
require "instrument_lib.pl";
require "mathlib.pl";

use Getopt::Long;
GetOptions(
    "root=s",    \$root,     "ext1=s",      \$out,
    "ext=s",     \$ext,      "dthresh=s",   \$dthresh,
    "mthresh=s", \$mthresh,  "radius=s",    \$radius,
    "ext_out=s", \$ext_out,  "boarder",     \$boarder,
    "small=s",   \$small,    "min_width=s", \$min_width,
    "log=s",     \$log,      "noclean",     \$noclean,
    "noreg",     \$noreg,    "noim",        \$noim,
    "norefine",  \$norefine, "nocheck",     \$nocheck,
    "instrum=s", \$instrum,  "debug",       \$debug,
    "bin=s",     \$bin
);

$debug = 1;
if (defined $debug) {
    $debug = 1;
}

# Read the programs paths
print "\nRead progs.ini \n";

if (exists $ENV{"PROGS_DIR"}) {
    $prog_dir = $ENV{"PROGS_DIR"} . "/";
} else {
    $prog_dir = "./";
}

#&read_progs(\%progs,"/home/phudelot/reduce_STRACK/progs.ini");
&read_progs(\%progs, $prog_dir . "progs.ini");

my @trash;

# PROGRAMS
$dfits            = $progs{"P_DFITS"};
$fitsort          = $progs{"P_FITSORT"};
$awk              = $progs{"P_GAWK"};
$gawk             = $progs{"P_GAWK"};
$sex              = $progs{"P_SEX"};
$stats            = $progs{"P_STATS"};
$ic               = $progs{"P_IC"};
$hough_weight_ref = $progs{"P_STRACK_WEIGHT"};
$filter_seg       = $progs{"P_STRACK_FILTER"};
$filter           = $progs{"P_STRACK_FILTER2"};
$getpix           = $progs{"P_GETPIX"};
$setpix           = $progs{"P_SETPIX"};
$cphead           = $progs{"P_CPHEAD"};

# CHECK RADIUS
$check_rad    = 60;                         # grouping tolerance radius
$dref         = $check_rad * $check_rad;    # radius (squared for check)
$thresh_check = 0.2;
$single_coeff = 1.5;     # dthresh * single_coeff = limit for smallflag
$lcoeff       = 0.5;     # lcoeff * Nx = limit length for lengthflag
$area_limit   = 1000;    # a track bigger than that is probably a star

#$area_limit = 5000;      # a track bigger than that is probably a star

if (!defined $min_width) {
    $min_width = 12;     # minimum width of a track (in pixels)
}
if (!defined $bin) {
    $bin = 1;
}
if (!defined $ext1) {
    $ext1 = "_SEG_hSN.fits";
}
if (!defined $dthresh) {
    $dthresh = 1.0;
}
if (!defined $mthresh) {
    $mthresh = 0.8;
}
if (!defined $ext_out) {
    $ext_out = "_flag.fits";
}
if (!defined $ext) {
    $ext = "OFCS";
}
if (defined $boarder) {
    $boarder1 = "$homedir/perl/VIRMOS/intermap1_CFH12K.fits";
}
if (!defined $log) {
    $log = $root . ".LOG";
}
if (!defined $instrum) {
    $instrum = "CFH12K";
}

$isat          = -1;
@sat_moseq     = ();
@sat_chip      = ();
@sat_crosschip = ();
@sat_number    = ();

# Get some parameters
my %par = ();
&get_param($instrum, \%par);

# Reference point
@ref = ($par{"REFX"}, $par{"REFY"});

# Chips layout
%layout = ();
get_layout($instrum, \%layout);

# loop on the chips
$image_ref = $root . "_1" . $ext . ".fits";
if (!-e $image_ref) {
    die "Missing $image_ref to get GABODSID \n";
}

$gabodsid = `$dfits $image_ref | $fitsort -d GABODSID | $awk '{print \$2}'`;
chomp $gabodsid;

# Get the list of chips
my %dat;
my $i;
&get_param($instrum, \%dat);
my @miss   = get_missing_chips($instrum);
my $nchips = $dat{"NCHIPS"};
for ($i = 1; $i <= $nchips; $i++) {
    foreach (@miss) {
        if ($_ == $i) {
            next;
        }
    }
    push(@lchips, $i);
}

$period = get_period($instrum, $gabodsid);

print "\nUsing chips : @lchips \n";

# --------- get BADCCD ----------
my @bad_ccd = ();
foreach (@lchips) {
    my $im   = $root . "_" . $_ . $ext . ".fits";
    my $test = `$dfits $im | $fitsort -d BADCCD | $awk '{print \$2}'`;
    chomp $test;
    if ($test !~ /KEY_N/ && $test == 1) {
        push(@bad_ccd, $_);
    }
}

print "\nBad CCDs : @bad_ccd \n";

###############################################################################
#########################  SINGLE chip STEP  ##################################
###############################################################################

print "\n\n ######### SINGLE CHIP STEP ############## \n\n";

foreach (@lchips) {
    $chip = $_;

    print "--- Working on chip $chip ... step1 \n";

    #die if ($chip == 5);

    $h1image = $root . "_" . $chip . $ext . $ext1;
    if (!(-e $h1image)) {
        die "$h1image or $h2image does not exist ... \n";
    }
    $Nx   = `$dfits $h1image | $fitsort -d NAXIS1 | $awk '{print \$2}'` + 0.0;
    $Ny   = `$dfits $h1image | $fitsort -d NAXIS2 | $awk '{print \$2}'` + 0.0;
    $refx = `$dfits $h1image | $fitsort -d REFX | $awk '{print \$2}'`;
    $refy = `$dfits $h1image | $fitsort -d REFY | $awk '{print \$2}'`;
    if ($refx !~ "KEY_N/A" && $refx ne "") {
        $refx += 0;
        $refy += 0;
    } else {
        $refx = 0;
        $refy = 0;
    }

    print "   Chip geometry : Nx/Ny $Nx $Ny refx/refy $refx $refy \n";

    $new_im = $h1image;

    # -------------  Run Sextractor  ---------------------------
    &write_param();

    my $cat   = $h1image . "_sex.cat";
    my $check = $h1image . "_check.fits";
    push(@trash, $cat);
    push(@trash, $check);

    if (defined $debug) {
        print "Run SExtractor on chip $chip \n";
        print
          "$sex -c default.sex_Z_$$ $new_im -CATALOG_NAME $cat -CHECKIMAGE_NAME $check -FILTER N \n";
    }
    system(
        "$sex -c default.sex_Z_$$ $new_im -CATALOG_NAME $cat -CHECKIMAGE_NAME $check -FILTER N \n"
    );

    $lcheck{$chip} = $check;

    # ---------- filter to get the detected objects  --------------
    my %data = ();
    &read_catc($cat, \%data);
    my @remove = ();
    my @keep   = ();
    &filter_detect(\%data, $dthresh, \@remove, \@keep);
    @all = (@remove, @keep);
    print "List of all detected tracks : @all \n";

    $drho = `$dfits $h1image | $fitsort -d DRHO | $awk '{print \$2}'` + 0.0;
    $dtheta =
      `$dfits $h1image | $fitsort -d DTHETA | $awk '{print \$2}'` + 0.0;

# ------------ filter to attribute the same segnumber to objets closer than ---------
    print
      "   Filter the objects to attribute the same segnumber to close objects ... \n";

    my @group  = ();
    my @listxy = ();
    my $flagg  = 0;
    my $i      = -1;
    foreach (@{$data{"NUMBER"}}) {
        $i++;
        my $xpeak = $data{"X_IMAGE"}->[$i];
        my $ypeak = $data{"Y_IMAGE"}->[$i];
        push(@listxy, [($xpeak, $ypeak)]);
    }
    $flagg = group_indexes(\@all, \@listxy, $check_rad, \@group);
    find_maxflux(\@group, \%data);

    if ($flagg > 0) {
        &filter_seg(\@group, $check, \%data)
          ;    # Change the hough sementation image according to the groups
        &filter_all(\@all, \@group)
          ;    # Change the all list according to the groups
        print "   New list of detected tracks : @all \n";
    }

    ############ FLAGS = vertical / horizontal ---> 10 degrees from vertical horizontal
    ############       = small
    ############       = crossed chip == 1

    # ----------- get the global equation for each ----------------------

    print "   Loop on detected tracks to get global equation \n";

    foreach (@all) {
        $il = $_;

        my @line_in =
          ($data{'X_IMAGE'}->[$_] - 1, $data{'Y_IMAGE'}->[$_] - 1);
        my $eq1 = ($data{'X_IMAGE'}->[$_] - 1) * $drho;
        my $eq2 = ($data{'Y_IMAGE'}->[$_] - 1 - $Ny / 2 + 0.5) * $dtheta;

        if (defined $debug) {
            print "Peak pixel $eq1 / $eq2 \n";
            print "eq1/eq2 $eq1 $eq2 \n";
        }

        # transform if ref is not (0,0)
        if ($refx != 0 || $refy != 0) {
            my @new_ref = (0, 0);
            my @old_ref = ($refx, $refy);
            my @eq      = ($eq1,  $eq2);
            &trans_ref(\@old_ref, \@new_ref, \@eq);
            $eq1 = $eq[0];
            $eq2 = $eq[1];
        }
        my @equ = ($eq1, $eq2);

        if (defined $debug) {
            print "Number "
              . $data{'NUMBER'}->[$_]
              . " with equation @equ \n";
        }

        my %limits_chip =
          get_chip_limits_instrum_ind0($period, $chip, $instrum);
        my %limits_mos = get_chippix_mosaic_instrum($period, $chip, $instrum);
        my @line_out = ();
        &trans_line(\%limits_chip, \%limits_mos, \@equ, \@line_out, \@ref);
        my @eq = @line_out;

        if (defined $debug) {
            print "Number "
              . $data{'NUMBER'}->[$_]
              . " with pix eq @equ mos equation @eq et ref @ref \n";
        }

        # ------- is a check needed --------
        print "       Get the warning flags ... \n";

        # ---- Vertical track (10 degree from vertical / horizontal)
        $vflag = 0;
        if (cos($eq2) > 0.985 || cos($eq2) < -0.985) {
            $vflag = 1;

            #print "Vertical flag \n";
        }
        if (cos($eq2) > -0.17 && cos($eq2) < 0.17) {
            $vflag = 2;

            #print "Horizontal flag \n";
        }

        # end ---- Vertical track

        # ---- Too small ?
        $sflag = 0;
        if ($data{'FLUX_MAX'}->[$_] < $dthresh * $single_coeff) {
            $sflag = 1;
        }

        # END ---- Too small ?

        # ----- Crossed chip == 1 ?
        $bflag = 0;

        # Get the list of crossed chips
        my @lcrossed = ();
        @lcrossed =
          check_linechip_instrum(\@line_out, \@ref, $instrum, $period);
        check_badccd(\@lcrossed, \@badccd);
        if (scalar(@lcrossed) == 1) {
            $bflag = 1;
        }
        if (defined $debug) {
            print "chip crossed = @lcrossed \n";
        }

        # END ---- Crossed chip == 1 ?

        # Max track length ------
        $lflag = 0;

        $llim = (min(($Nx, $Ny)) * $lcoeff) * (min(($Nx, $Ny)) * $lcoeff);

        my @PP1;
        my @PP2;
        get_point_chip(\@equ, \%limits_chip, \@PP1, \@PP2);
        my $length2 =
          (($PP1[0] - $PP2[0]) * ($PP1[0] - $PP2[0]) +
              ($PP1[1] - $PP2[1]) * ($PP1[1] - $PP2[1]));

        if ($length2 < $llim) {
            $lflag = 1;
        } else {
            $lflag = 0;
        }

        # END --- Max track length

        # area limit ----------
        if ($data{'ISOAREA_IMAGE'}->[$il] > $area_limit) {
            $aflag = 1;
        } else {
            $aflag = 0;
        }

        # END------- area limit

        # Save parameters for global check
        @minmax = (
            $data{'XMIN_IMAGE'}->[$il] - 20,
            $data{'XMAX_IMAGE'}->[$il] + 20,
            $data{'YMIN_IMAGE'}->[$il] - 20,
            $data{'YMAX_IMAGE'}->[$il] + 20
        );

        my @flags = ($sflag, $vflag, $bflag, $lflag, $aflag);
        print
          "       Final flags : small/vertical/boarder/length/area : @flags \n";

        push(@sat_number,  $data{'NUMBER'}->[$il]);
        push(@sat_area,    $data{'ISOAREA_IMAGE'}->[$il]);
        push(@sat_chip,    $chip);
        push(@sat_checkim, $check);
        push(@sat_minmax,    [@minmax]);
        push(@sat_moseq,     [@eq]);
        push(@sat_moseq2,    [($eq[0] / $drho, $eq[1] / $dtheta)]);
        push(@sat_chipeq,    [@line_in]);
        push(@sat_crosschip, [@lcrossed]);
        push(@sat_flag,      [@flags]);

        push(@sat_ref, [@ref]);
        $refchip{$chip}->[0] = $refx;
        $refchip{$chip}->[1] = $refy;

        push(@sat_dd, [($drho, $dtheta)]);
        push(@sat_NN, [($Nx,   $Ny)]);

        $isat++;
    }
    print "\n\n";
}

#die;

###############################################################################
#########################  GLOBAL check STEP ##################################
###############################################################################

print "\n\n ######### GLOBAL MOSAIC STEP ############## \n\n";

# log file header
open(LOG, ">$log");
print LOG <<END;
# 1 CHIP
# 2 NUMBER
# 3 MOS_HOUGH_R
# 4 MOS_HOUGH_THETA
# 3 CHIP_HOUGH_R
# 4 CHIP_HOUGH_THETA
# 5 BOARDER_FLAG 
# 6 VERTICAL_FLAG
# 7 SMALL_FLAG
END
close LOG;

# Debug log file
if (defined $debug) {
    $ii = -1;
    foreach (@sat_number) {
        $ii++;
        push(@x, $sat_moseq[$ii]->[0]);
        push(@y, $sat_moseq[$ii]->[1]);
        $chip = $sat_chip[$ii];
        system( "echo "
              . $sat_moseq[$ii]->[0] . " "
              . $sat_moseq[$ii]->[1]
              . " $chip  >> log_$$ \n");
    }
}

########## GROUP THE MOSAIC EQUATIONS

print "Grouping the tracks \n";

@indexes_satmos = ();
my $ij = -1;
foreach (@sat_number) {
    $ij++;
    push(@indexes_satmos, $ij);
}

@group_mos = ();

print "label1 \n";
group_indexes_px_complete(
    \@indexes_satmos, \@sat_moseq2, \@sat_checkim,
    \@sat_number,     $check_rad,   \@group_mos
);

print "Groups of tracks :\n";
foreach (@group_mos) {
    print "  New group (chips): ";
    foreach (@{$_->{"L"}}) {
        $chip = $sat_chip[$_];
        print " $chip ";
    }
    print "\n";
    print "      Flags :";
    foreach (@{$_->{"L"}}) {
        @chip = @{$sat_flag[$_]};
        print " @chip ";
    }
    print "\n";
    print "      Indexes :";
    foreach (@{$_->{"L"}}) {
        print " $_ ";
    }
    print "\n";

}

######################################################
########## Loop on the detected tracks ###############
###### Check vertical / small / other chip ###########
######### ONLY SINGLE object groups ##################
######################################################

print "\n\n ######### GLOBAL TRACK CHECK ############## \n\n";

@bad_strack  = ();
@good_strack = ();

print "Loop on the groups \n";
my $igr = -1;
foreach (@group_mos) {
    $igr++;
    my $ig = $_;

    print "  Working on group $igr \n";

    if (defined $debug) {
        print "New group : ";
        foreach (@{$_->{"L"}}) {
            $chip = $sat_chip[$_];
            print " $chip ";
        }
        print "\n";
        foreach (@{$_->{"L"}}) {
            @chip = @{$sat_flag[$_]};
            print " @chip ";
        }
        print "\n";
    }

    if (scalar(@{$ig->{"L"}}) > 1) {

        # At least from two different chips
        my @diff  = ();
        my @diff2 = ();
        foreach (@{$ig->{"L"}}) {
            if (!is_in_num(\@diff, $sat_chip[$_])) {
                push(@diff,  $sat_chip[$_]);
                push(@diff2, $_);
            }
        }
        if (scalar(@diff) > 1) {
            foreach (@{$ig->{"L"}}) {
                push(@good_strack, $_);
            }

            # Check layout if two vertical flagged tracks
            my $fv = 0;
            if (   scalar(@diff2) == 2
                && $sat_flag[$diff2[0]]->[1] > 0
                && $sat_flag[$diff2[1]]->[1] > 0)
            {
                $fv = check_vertical_layout($ig, \%layout, \@sat_flag);
            }
            if ($fv == 0) {
                print "    Multichip group ... valid \n";
                next;
            }
        }
    }
    print "    Singlechip group ... check needed (loop on tracks)\n";

    # Single or more than one but on one chip

    foreach (@{$ig->{"L"}}) {
        my $ind = $_;

        # Single object group
        @flags = @{$sat_flag[$ind]};
        my ($sflag, $vflag, $bflag, $lflag, $aflag) = @flags;

        print
          "        Object $ind : flags = $sflag,$vflag,$bflag,$lflag,$aflag \n";

        my $rem  = 0;
        my $safe = 0;

        # Boarder ?
        if (($bflag && $sflag) || ($bflag && $lflag)) {
            $rem = 1;
            print "Boarder flag !!\n";
        } elsif ($bflag == 1) {
            $safe = 1;
            print "Boarder flag !!\n";
        } else {
            $rem = 1;
        }

        # vertical OR small/length or TOO BIG
        if ($sflag || $vflag || $lflag || $aflag) {
            print "Vertical/Small flag !!\n";
            $rem = 1;
        }

        # Remove bad tracks
        if ($rem == 1) {
            print "splice it ... \n";
            push(@to_splice, $igr);

            # Add to bad_tracks
            print "          -> bad $ind chip : "
              . $sat_chip[$ind]
              . " segnumber "
              . $sat_number[$ind] . "\n";
            push(@bad_strack, $ind);
        } else {
            print "          -> good $ind chip : "
              . $sat_chip[$ind]
              . " segnumber "
              . $sat_number[$ind] . "\n";
            push(@good_strack, $ind);
        }

    }
}

print "\n\n Groups to splice : @to_splice \n";

$igg = -1;
foreach (@group_mos) {
    $igg++;
    if (!is_in_num(\@to_splice, $igg)) {
        push(@groups, $_);
    }
}

#die;

####################################################
####### Check for undetected chips #################
####################################################

print "\n\nNew groups of tracks :\n";
foreach (@groups) {
    print "  New group (chips): ";
    foreach (@{$_->{"L"}}) {
        $chip = $sat_chip[$_];
        print " $chip ";
    }
    print "\n";
    print "      Flags :";
    foreach (@{$_->{"L"}}) {
        @chip = @{$sat_flag[$_]};
        print " @chip ";
    }
    print "\n";
    print "      Indexes :";
    foreach (@{$_->{"L"}}) {
        print " $_ ";
    }
    print "\n";

}

# debug print
if (defined $debug) {
    foreach (@groups) {
        @center = @{$_->{'C'}};
        @list   = @{$_->{'L'}};
        @disp   = @{$_->{'D'}};
        print "center @center and list @list and disp @disp \n";
        print "chips : ";
        foreach (@{$_->{'L'}}) {
            print " " . $sat_chip[$_] . " ";
        }
        print "\n";
    }
}

##########################################################
# Keep groups with more than one member + check linechips#
##########################################################

print "\n\n ######### CHECK FOR UNDETECTED TRACKS ############## \n\n";
print "Loop on the groups : \n";

$nflag = 0;
my $iig = -1;
foreach (@groups) {
    $iig++;
    print " Working on group $iig :\n";
    $ig = $_;
    if (scalar(@{$_->{'L'}}) > 1) {    # Multiobject group
        print "    -> Multiobject group \n";

# Loop on the tracks (each chip has to be on each track of the group) -----------------
        my @cross = @{$sat_crosschip[$ig->{'L'}->[0]]};

        foreach (@{$ig->{'L'}}) {

            # Get the crossed chips
            my @lcross = @{$sat_crosschip[$_]};

            my @new_cross;
            foreach (@cross) {
                if (is_in_num(\@lcross, $_) == 1) {
                    push(@new_cross, $_);
                }
            }
            @cross = ();
            @cross = @new_cross;
        }

        print "    Crossed chips : @cross \n";

        # Get list of chips
        my @chipss = ();
        foreach (@{$ig->{'L'}}) {
            push(@chipss, $sat_chip[$_]);
        }

# compare to the group members --------(expected chip crossed VS detected chips)---------------
        foreach (@cross) {
            $wchip = $_;

            # -------------------               --------------------
            ############ MISSING CHIP FOUND ########################
            # -------------------               --------------------
            if (!is_in_num(\@chipss, $wchip)) {    # Missing track on chip
                print "     Missing $wchip in @chipss \n";

             # --------- mosaic equation of the center of the group ----------
             # refine the center (sigclip if more than 4 members)
                my $eq;
                my $sig = 2;
                my @eq_px =
                  center_group_sig(\@sat_moseq2, $ig->{'L'}, $ig->{'C'},
                    $ig->{'D'}, $sig);
                @eq = ($eq_px[0] * $drho, $eq_px[1] * $dtheta);

                my @eq2_px = @{$ig->{'C'}};
                my @eq2 = ($eq2_px[0] * $drho, $eq2_px[1] * $dtheta);

                if (defined $debug) {
                    print "global missing track equation @eq and not @eq2 \n";
                }

                # --------- Transform to Chip equation ---------------
                my @ceq = ();
                my %limits_chip =
                  get_chip_limits_instrum_ind0($period, $wchip, $instrum);
                my %limits_mos =
                  get_chippix_mosaic_instrum($period, $wchip, $instrum);
                &trans_line2(\%limits_chip, \%limits_mos, \@eq, \@ceq, \@ref);

                #print "    Mos equation in missing chip @ceq \n";

                # ----------- Find closest chips ----------
                #my @cchip = get_closest_chips(\@chipss,$wchip);
                my @cchip = get_closest_chips(\@chipss, $wchip, \@eq);
                my @cnumber1;
                my @cnumber2;
                print "        Closest chips = @cchip \n";

                # get line seg number for the two closest chips
                foreach (@{$ig->{'L'}}) {
                    if ($sat_chip[$_] == $cchip[0]) {
                        $number = $sat_number[$_];
                        push(@cnumber1, $number);
                    }
                    if ($sat_chip[$_] == $cchip[1]) {
                        $number = $sat_number[$_];
                        push(@cnumber2, $number);
                    }
                }

                if (defined $debug) {
                    print "chip " . $cchip[0] . " segnum @cnumber1 \n";
                    print "chip " . $cchip[1] . " segnum @cnumber2 \n";
                }

                #die;

                ##############################
                ####### Get the width ########
                ##############################

                print "          Get Width of undetected chip ... \n";

                # Width on chip1
                my @beq = ();
                $new_im = $lcheck{$cchip[0]};
                %limits_mos =
                  get_chippix_mosaic_instrum($period, $cchip[0], $instrum);
                &trans_line2(\%limits_chip, \%limits_mos, \@eq, \@beq, \@ref);
                if (defined $debug) {

                    foreach (keys(%limits_mos)) {
                        print "test lim_mos $_ " . $limits_mos{$_} . "\n";
                    }

                    print "ref = @ref \n";
                    print
                      "get width from $new_im and equatiob @eq ... chip = "
                      . $cchip[0] . "\n";
                }
                my $width1 =
                  get_width($new_im, \@cnumber1, \@beq, \%limits_chip);
                if (defined $debug) {
                    print "width1 = $width1\n";
                }

                # width on chip2
                my @deq = ();
                $new_im = $lcheck{$cchip[1]};
                %limits_mos =
                  get_chippix_mosaic_instrum($period, $cchip[1], $instrum);
                &trans_line2(\%limits_chip, \%limits_mos, \@eq, \@deq, \@ref);
                if (defined $debug) {

                    foreach (keys(%limits_mos)) {
                        print "test lim_mos $_ " . $limits_mos{$_} . "\n";
                    }

                    print "ref = @ref \n";
                    print
                      "get width from $new_im and equatiob @eq ... chip = "
                      . $cchip[1] . "\n";
                }
                my $width2 =
                  get_width($new_im, \@cnumber2, \@deq, \%limits_chip);
                if (defined $debug) {
                    print "width2 = $width2\n";
                }

		# treat NaNs in the 'width' variables:
                if (isnan($width1) || isnan($width2)) {
                    if (isnan($width1) && !isnan($width2)) {
                        $width = $width2;
                    } else {
                        if (!isnan($width1) && isnan($width2)) {
                            $width = $width1;
                        } else {
                            if (isnan($width1) && isnan($width2)) {
                                $width = $min_width;
                            } else {
                                $width = max($width1, $width2);
                            }
                        }
                    }
                }                                

                # ensure minimum width:
                $width = max($width, $min_width);

                print "          Final width = $width \n";

                # Try to refine the position of the line
                print "          Line equation before refinement"
                  . $ceq[0] . "  "
                  . $ceq[1]
                  . " width "
                  . $width . "  \n";
                my $flag = 0;
                if (!(defined $norefine && defined $nocheck)) {

                 # Get a subimage around the line in the hough transform image
                    my $x = $ceq[0] / $drho;
                    my $y;
                    if ($ceq[1] > 3.141592654) {
                        $y -= 2 * 3.141592654;
                    }
                    $y = ($ceq[1] + 3.141592654) / $dtheta;

                    my $y = ($ceq[1] + 3.141592654) / $dtheta;

                    my $xmin = int($x - 50 + 0.5);
                    my $xmax = int($x + 50 + 0.5);
                    my $ymin = int($y - 50 + 0.5);
                    my $ymax = int($y + 50 + 0.5);

                    if ($xmin < 1) {
                        $xmin = 1;
                    }
                    if ($ymin < 1) {
                        $ymin = 1;
                    }
                    if ($xmax > $Nx) {
                        $xmax = $Nx;
                    }
                    if ($ymax > $Ny) {
                        $ymax = $Ny;
                    }

                    #print "test $Nx $Ny \n";

                    # Extract the subimage
                    my $h1image = $root . "_" . $wchip . $ext . $ext1;
                    $him = $h1image;

                    my $im = "test.ZZ.fits";
                    push(@trash, $im);

                    $dx = abs($xmax - $xmin) + 1;
                    $dy = abs($ymax - $ymin) + 1;
                    if ($ymax < $ymin) {
                        $ymin = $ymax;
                    }
                    if ($xmax < $xmin) {
                        $xmin = $xmax;
                    }
                    system(
                        "cat $him | makesubimage $xmin $ymin $dx $dy > $im \n"
                    );

                #print "cat $him | makesubimage $xmin $ymin $dx $dy > $im \n";

                    # Rerun Sextractor with lower threshold
                    &write_param();

                    my $cat = $h1image . "_sex.cat.Z";
                    push(@trash, $cat);
                    system("echo X_IMAGE > default.param_Z_$$");
                    system("echo Y_IMAGE >> default.param_Z_$$");

                    system(
                        "$sex -c default.sex_Z_$$ $im -CATALOG_NAME $cat -CHECKIMAGE_TYPE NONE -DETECT_THRESH $thresh_check -ANALYSIS_THRESH $thresh_check\n"
                    );

                    # Find the center of the line
                    open(IN, $cat);
                    $flag = 0;
                    while (defined($oneline = <IN>)) {
                        next if $oneline =~ /\s*\#/;
                        $oneline =~ /(\S+)\s+(\S+)/;
                        my $x0 = $1 + $xmin - 2;
                        my $y0 = $2 + $ymin - 2;
                        my $d =
                          ($x - $x0) * ($x - $x0) + ($y - $y0) * ($y - $y0);

                        if ($d < $check_rad * $check_rad && $d < $dref) {
                            $flag = 1;
                            if (!defined $norefine) {
                                $ceq[0] = $x0 * $drho;
                                $ceq[1] = $y0 * $dtheta - 3.1415927;
                            }
                        }

                    }
                    print
                      "          Line equation after refinement : @ceq \n";
                }

                # die;

                # Save the track if contiguous to a chip where it was detected
                if (!defined $nocheck && $flag == 0) {
                    my @chi = get_closest_chips_check(\@cross, $wchip, \@eq);

                    if (   is_in_num(\@chipss, $chi[0])
                        || is_in_num(\@chipss, $chi[1]))
                    {
                        $flag = 1;
                    }
                }

                # build the flagmask
                if (defined $nocheck || $flag == 1) {
                    push(
                        @{$flag[$nflag]->{$wchip}},
                        [($ceq[0], $ceq[1], $width)]
                    );
                    $nflag++;
                }
            }
        }

    } else {
        print "    -> Singleobject group \n";

        # Shall we consider undeteced tracks ??????
        if (1) {
            my @cross = @{$sat_crosschip[$ig->{'L'}->[0]]};
            print "crossed @cross \n";

            # Keep only if it crosses only one more chips
            if (scalar(@cross) == 2) {
                if ($cross[0] == $sat_chip[$ig->{'L'}->[0]]) {
                    $wchip = $cross[1];
                    $ochip = $cross[0];
                } else {
                    $wchip = $cross[0];
                    $ochip = $cross[1];
                }
                print "Missing chip $missing_chip \n";

                # Mos equation
                my @eq = @{$sat_moseq[$ig->{'L'}->[0]]};
                print "equation eq @eq \n";

                # Width on chip1
                my @beq = ();
                $new_im = $lcheck{$ochip};
                my %limits_mos =
                  get_chippix_mosaic_instrum($period, $ochip, $instrum);
                my %limits_chip =
                  get_chip_limits_instrum_ind0($period, $ochip, $instrum);
                &trans_line2(\%limits_chip, \%limits_mos, \@eq, \@beq, \@ref);
                print "equation beq @beq \n";

                if (1) {

                    foreach (keys(%limits_mos)) {
                        print "test lim_mos $_ " . $limits_mos{$_} . "\n";
                    }

                    print "ref = @ref \n";
                    print
                      "get width from $new_im and equatiob @eq ... chip = "
                      . $ochip . "\n";
                }

                my @cnumber1 = ($sat_number[$ig->{'L'}->[0]]);
                my $width1 =
                  get_width($new_im, \@cnumber1, \@beq, \%limits_chip);

                if (1) {
                    print "width1 = $width1\n";
                }

                # Try to refine the position of the line
                print "          Line equation before refinement"
                  . $ceq[0] . "  "
                  . $ceq[1]
                  . " width "
                  . $width . "  \n";
                my $flag = 0;

                if (!(defined $norefine && defined $nocheck)) {

                    my %limits_mos =
                      get_chippix_mosaic_instrum($period, $wchip, $instrum);
                    my %limits_chip =
                      get_chip_limits_instrum_ind0($period, $wchip, $instrum);
                    &trans_line2(\%limits_chip, \%limits_mos, \@eq, \@ceq,
                        \@ref);
                    my $x = $ceq[0] / $drho;
                    my $y;
                    if ($ceq[1] > 3.141592654) {
                        $y -= 2 * 3.141592654;
                    }
                    my $y = ($ceq[1] + 3.141592654) / $dtheta;

                    my $xmin = int($x - 50 + 0.5);
                    my $xmax = int($x + 50 + 0.5);
                    my $ymin = int($y - 50 + 0.5);
                    my $ymax = int($y + 50 + 0.5);

                    if ($xmin < 1) {
                        $xmin = 1;
                    }
                    if ($ymin < 1) {
                        $ymin = 1;
                    }
                    if ($xmax > $Nx) {
                        $xmax = $Nx;
                    }
                    if ($ymax > $Ny) {
                        $ymax = $Ny;
                    }

                    # Extract the subimage
                    my $h1image = $root . "_" . $wchip . $ext . $ext1;
                    $him = $h1image;

                    my $im = "test.ZZ.fits";
                    push(@trash, $im);

                    $dx = abs($xmax - $xmin) + 1;
                    $dy = abs($ymax - $ymin) + 1;
                    if ($ymax < $ymin) {
                        $ymin = $ymax;
                    }
                    if ($xmax < $xmin) {
                        $xmin = $xmax;
                    }
                    system(
                        "cat $him | makesubimage $xmin $ymin $dx $dy > $im \n"
                    );

                    # Rerun Sextractor with lower threshold
                    &write_param();

                    my $cat = $h1image . "_sex.cat.Z_$$";
                    push(@trash, $cat);
                    system("echo X_IMAGE > default.param_Z_$$");
                    system("echo Y_IMAGE >> default.param_Z_$$");

                    system(
                        "$sex -c default.sex_Z_$$ $im -CATALOG_NAME $cat -CHECKIMAGE_TYPE NONE -DETECT_THRESH $thresh_check -ANALYSIS_THRESH $thresh_check\n"
                    );

                    # Find the center of the line
                    open(IN, $cat);
                    $flag = 0;
                    while (defined($oneline = <IN>)) {
                        next if $oneline =~ /\s*\#/;
                        $oneline =~ /(\S+)\s+(\S+)/;
                        my $x0 = $1 + $xmin - 2;
                        my $y0 = $2 + $ymin - 2;
                        my $d =
                          ($x - $x0) * ($x - $x0) + ($y - $y0) * ($y - $y0);

                        if ($d < $check_rad * $check_rad && $d < $dref) {
                            $flag = 1;
                            if (!defined $norefine) {
                                $ceq[0] = $x0 * $drho;
                                $ceq[1] = $y0 * $dtheta - 3.1415927;
                            }
                        }
                    }
                    print
                      "          Line equation after refinement : @ceq \n";
                }

                # build the flagmask
                push(
                    @{$flag[$nflag]->{$wchip}},
                    [($ceq[0], $ceq[1], $width1)]
                );
                $nflag++;
                print "Add line "
                  . $ceq[0] . "  "
                  . $ceq[1] . " "
                  . $width1 . "\n";
            }
        }

    }
    print "\n";
}

#die;
########### DEBUG ############
#$debug = 1;
if ($debug) {
    foreach (@good_strack) {
        my $ii = $_;
        push(@x, $sat_moseq[$ii]->[0]);
        push(@y, $sat_moseq[$ii]->[1]);
        $chip = $sat_chip[$ii];
        system( "echo "
              . $sat_moseq[$ii]->[0] . " "
              . $sat_moseq[$ii]->[1]
              . " $chip  >> log2_$$ \n");
    }
}

if ($debug) {
    foreach (@bad_strack) {
        my $ii = $_;
        push(@x, $sat_moseq[$ii]->[0]);
        push(@y, $sat_moseq[$ii]->[1]);
        $chip = $sat_chip[$ii];
        system( "echo "
              . $sat_moseq[$ii]->[0] . " "
              . $sat_moseq[$ii]->[1]
              . " $chip  >> log2_bad_$$ \n");
    }
}

#$debug = 0;

######## DEBUG #############

##########################################################
############### Create flag images #######################
##########################################################

print "\n\n ######### CREATE THE MASKS ############## \n\n";
print "Loop on the chips : \n";

# Create flag images
foreach (@lchips) {
    print "\n\n Working on chip $_ \n";

    $chip = $_;

    my $reffx   = $refchip{$chip}->[0];
    my $reffy   = $refchip{$chip}->[1];
    my @old_ref = ($reffx, $reffy);

    # Remove false detection from the segmentation image
    if (scalar(@bad_strack) > 0) {
        print "   Remove bad tracks from the hough image \n";

        $badcat = "badcat" . $chip . ".Z_$$";
        push(@trash, $badcat);

        if (0) {

            # print list of bad tracks
            open(OUT, ">" . $badcat);
            foreach (@bad_strack) {
                next if ($chip != $sat_chip[$_]);

                $number = $sat_number[$_];
                print OUT "$number "
                  . $sat_minmax[$_]->[0] . " "
                  . $sat_minmax[$_]->[2] . " "
                  . $sat_minmax[$_]->[1] . " "
                  . $sat_minmax[$_]->[3] . "\n";
            }
            close OUT;

            # Do it
            $ref = $lcheck{$chip};
            $fin = "check_fin_" . $chip . $$ . ".fits";
            system("$filter_seg $ref $badcat $fin \n");
        } else {

            # print list of bad tracks
            open(OUT, ">" . $badcat);
            my $fflag = 0;
            foreach (@bad_strack) {
                next if ($chip != $sat_chip[$_]);

                $number = $sat_number[$_];
                print OUT "$number 0 \n";
                $fflag = 1;
            }
            close OUT;

            # Do it
            if ($fflag == 1) {
                $ref = $lcheck{$chip};
                $fin = "check_fin_" . $chip . $$ . ".fits";
                print "       Filter the $ref file into $fin with $badcat \n";
                print("$filter -in $fin -cat $badcat \n");
                system("cp $ref $fin \n");
                system("$filter -in $fin -cat $badcat \n");
                print "done \n";
                push(@trash, $fin);
            } else {
                $fin = $lcheck{$chip};
            }

        }
    } else {
        $fin = $lcheck{$chip};
    }

    $final1 = $root . "_" . $chip . $ext . "_h1_SEG.fits";
    system("$ic '\%1 0.5 0 1 < ?' $fin > $final1 \n");
    push(@trash, $final1);

    # Add drho / dtheta to the headers
    # system("cphead $h1image $final1 DRHO DTHETA \n");
    system("echo DRHO > listk.Z_$$ \n");
    system("echo DTHETA >> listk.Z_$$ \n");
    system("$cphead $h1image $final1  \@listk.Z_$$ \n");
    system("rm -f listk.Z_$$ \n");

    if (!defined $noim) {
        print "      Create mask images \n";

        # Inverse transform the segmentation images
        $final = $root . "_" . $chip . $ext . "_flag.fits";
        system("$hough2_1im $final1 $image_ref $final 0.5 ");

        #  --------- flag the additionnal satellite tracks ---------
        print "      Flag the additionnal tracks $chip \n";
        foreach (@flag) {
            if (exists($_->{$chip})) {
                @list = @{$_->{$chip}};
                &add_flag($final, \@list);
            }
        }

        if (defined $radius && $radius > 1) {
            &grow_fits($final, $radius);
        }
    }

    if ($debug) {
        print "Before reg $final1 \n";
    }

    #########################################################
    ################# Build the reg files ###################
    #########################################################
    if (!defined $noreg) {
        print "     Create the reg files \n";

        $regfile = $root . "_" . $chip . $ext . ".reg";
        open(REG, ">$regfile");
        print REG<<END;
\# \# Region file format: DS9 version 3.0
\# \# Filename: $root."_".$chip.$ext."fits";
\# global color=green font="helvetica 10 normal" select=1 edit=1 move=1 delete=1 include=1 fixed=0 source
END

        # Get the list of pixels for each track
        $final1 = $fin;

        @list = `$getpix -in $final1 -vmin 0.5  \n`;
        my %data = ();
        foreach (@list) {
            my $line = $_;
            if ($line =~ /(\d+),(\d+).*=\s(\S+)/) {
                push(@{$data{$3}}, [($1, $2)]);
            }
        }
        @keep = keys(%data);

        if ($debug) {
            print "keep = @keep \n";
        }

        # loop on the tracks
        foreach (@keep) {
            $segn = $_;
            my @list2 = @{$data{$_}};
            my @pix;

            # substract 1 to get the index0 and convert to r,theta
            foreach (@list2) {
                my $eq1 = ($_->[0] - 1) * $drho;
                my $eq2 = ($_->[1] - 1 - $Ny / 2 + 0.5) * $dtheta;

                # transform eq if ref not 0,0
                if ($reffx != 0 || $reffy != 0) {
                    my @eq = ($eq1, $eq2);
                    my @new_ref = (0, 0);
                    &trans_ref(\@old_ref, \@new_ref, \@eq);
                    ($eq1, $eq2) = @eq;
                }
                push(@pix, [($eq1, $eq2)]);

            }
            $npx = scalar(@pix);

            # get the equation of the peak
            my $ii  = -1;
            my @ceq = ();
            foreach (@sat_moseq) {
                $ii++;
                if ($debug) {
                    print "segn $segn and chip $chip \n";
                    print "segn "
                      . $sat_number[$ii]
                      . " chip "
                      . $sat_chip[$ii] . "\n";
                }

                if ($sat_chip[$ii] == $chip && $sat_number[$ii] == $segn) {
                    @ceq = @$_;
                    last;
                }
            }

            # get the intersection points with the edges (limits)
            my @listp;
            %limits = get_chip_limits_instrum_ind0($period, $chip, $instrum);
            my @side = ();
            &get_limits_inter(\@pix, \%limits, \@listp, \@side);

            if ($debug) {
                foreach (@listp) {
                    my @tmp = @$_;
                    print "point @tmp \n";
                }
            }

            # minimum size for the track
            min_size(\@listp, \@side, $min_width);

            # Write the reg file
            my $it = -1;
            foreach (@listp) {
                $it++;
                $listp[$it]->[0] = $listp[$it]->[0] * 1.0;
                $listp[$it]->[1] = $listp[$it]->[1] * 1.0;
            }

            print REG "POLYGON(";
            $fp = shift @listp;
            print REG $fp->[0] . "," . $fp->[1];
            foreach (@listp) {
                print REG "," . $_->[0] . "," . $_->[1];
            }
            print REG ")\n";
        }

        # add globally detected tracks
        my %limits_chip = get_chip_limits_instrum($period, $chip, $instrum);
        my $sizex       = $limits_chip{"X_X_Y"};
        my $sizey       = $limits_chip{"Y_X_Y"};
        my @sizee       = ($sizex, $sizey);
        foreach (@flag) {
            if (exists($_->{$chip})) {
                @list = @{$_->{$chip}};
                &add_flag_reg($final, \@list, \@sizee);
            }
        }

        close REG;

        # Remove empty reg files
        my $nlines =
          `cat $regfile | $gawk '\$1 != "#" {print \$0}' | wc | $gawk '{print \$1}'`;
        chomp $nlines;
        if ($nlines == 0) {
            system("rm -f $regfile \n");
        }
    }
}

# add not needed temporary files to the trash can:
push(
    @trash,
    (
        "default.sex_Z_$$", "default.param_Z_$$",
        "default.nnw_Z_$$", "default.conv_Z_$$"
    )
);
push(@trash, "tmp_flag.ZZ_$$");
push(@trash, ("log_$$", "log2_$$", "log2_bad_$$"));

if (!defined $noclean) {
    my @listt = ($f1, $f2, $final1, $final2, $fin);
    foreach (@listt) {
        if (-e "$_") {
            system("rm -f $_ \n");
        }
    }
    foreach (@trash) {
        if (-e "$_") {
            system("rm -f $_ \n");
        }
    }
}

############################################################################################################
sub center_group_sig {
    my ($leq, $lnum, $cen, $disp, $sig) = @_;

    $l1m = $cen->[0] - $disp->[0] * $sig;
    $l1M = $cen->[0] + $disp->[0] * $sig;
    $l2m = $cen->[1] - $disp->[1] * $sig;
    $l2M = $cen->[1] + $disp->[1] * $sig;

    if (defined $debug) {
        print "limits $l1m $l1M $l2m $l2M \n";
    }

    my $tmp1 = 0;
    my $tmp2 = 0;
    my $n    = 0;
    foreach (@$lnum) {
        my $eq1 = $leq->[$_]->[0];
        my $eq2 = $leq->[$_]->[1];

        if (defined $debug) {
            print "eq $eq1 $eq2  \n";
        }

        if ($eq1 >= $l1m && $eq1 <= $l1M && $eq2 >= $l2m && $eq2 <= $l2M) {
            $tmp1 += $eq1;
            $tmp2 += $eq2;
            $n++;
        }
    }
    return ($tmp1 / $n, $tmp2 / $n);
}

############################################################################################################
sub add_flag_reg {
    my ($image, $list, $size) = @_;
    my $Nx = $size->[0];
    my $Ny = $size->[1];

    foreach (@$list) {
        my $r      = $_->[0];
        my $theta  = $_->[1] - 3.1415927 / 2;
        my $thetar = $_->[1];
        my $width  = $_->[2];

	if ($width > 0.0) {
            # Mask lines
            if (abs(cos($theta)) > 0.5) {    # y = ax + b
                $ydec = ($width / 2.0) / abs(cos($theta));
                $a    = -cos($thetar) / sin($thetar);
                $b    = $r / sin($thetar);
                $x1   = 0;
                $y1   = $a * $x1 + $b;
                $x2   = $Nx - 1;
                $y2   = $a * $x2 + $b;
    
                $x1m = 0;
                $y1m = $y1 - $ydec;
                $x1M = 0;
                $y1M = $y1 + $ydec;
                $x2m = $x2;
                $y2m = $y2 - $ydec;
                $x2M = $x2;
                $y2M = $y2 + $ydec;
    
                my @listp = ();
                push(@listp, [($x1m, $y1m)]);
                push(@listp, [($x1M, $y1M)]);
                push(@listp, [($x2m, $y2m)]);
                push(@listp, [($x2M, $y2M)]);
                my @side = (2, 4);
                &min_size(\@listp, \@side, $min_width);
                $x1m = $listp[0]->[0];
                $y1m = $listp[0]->[1];
                $x1M = $listp[1]->[0];
                $y1M = $listp[1]->[1];
                $x2m = $listp[2]->[0];
                $y2m = $listp[2]->[1];
                $x2M = $listp[3]->[0];
                $y2M = $listp[3]->[1];
            } else {    # x = ay + b
                $xdec = ($width / 2.0) / abs(sin($theta));
                $a    = -sin($thetar) / cos($thetar);
                $b    = $r / cos($thetar);
    
                $y1 = 0;
                $x1 = $b;
                $y2 = $Ny - 1;
                $x2 = $a * $y2 + $b;
    
                $x1m = $x1 - $xdec;
                $y1m = 0;
                $x1M = $x1 + $xdec;
                $y1M = 0;
                $x2m = $x2 - $xdec;
                $y2m = $y2;
                $x2M = $x2 + $xdec;
                $y2M = $y2;
    
                my @listp = ();
                push(@listp, [($x1m, $y1m)]);
                push(@listp, [($x1M, $y1M)]);
                push(@listp, [($x2m, $y2m)]);
                push(@listp, [($x2M, $y2M)]);
                my @side = (1, 3);
                &min_size(\@listp, \@side, $min_width);
                $x1m = $listp[0]->[0];
                $y1m = $listp[0]->[1];
                $x1M = $listp[1]->[0];
                $y1M = $listp[1]->[1];
                $x2m = $listp[2]->[0];
                $y2m = $listp[2]->[1];
                $x2M = $listp[3]->[0];
                $y2M = $listp[3]->[1];
    
            }
    
            # Convert to float
            $x1m *= 1.0;
            $x1M *= 1.0;
            $y1m *= 1.0;
            $y1M *= 1.0;
            $x2m *= 1.0;
            $x2M *= 1.0;
            $y2m *= 1.0;
            $y2M *= 1.0;
    
            # order it
            my (@pp11, @pp12, @pp21, @pp22);
            my @p1 = ($x1m, $y1m);
            my @p2 = ($x1M, $y1M);
            my @p3 = ($x2m, $y2m);
            my @p4 = ($x2M, $y2M);
            &get_point_order(\@p1, \@p2, \@p3, \@p4, \@pp11, \@pp12, \@pp21,
                \@pp22);
    
            print REG
              "POLYGON($pp11[0],$pp11[1],$pp12[0],$pp12[1],$pp21[0],$pp21[1],$pp22[0],$pp22[1])\n";
        }
    }
}
#####################################################################
sub get_limits_inter {
    my ($pix, $limits, $listp, $side) = @_;

    my @side;

    # first line to get the sides
    my @pixx = @$pix;
    $fpix = $pixx[0];
    my $theta = $fpix->[1];
    my @side_ref;
    if (abs(cos($theta)) > 0.5) {
        @side_ref = (1, 3);
    } else {
        @side_ref = (2, 4);
    }

    foreach (@side_ref) {
        push(@$side, $_);
    }

    # get the intersection points
    my @pix1X = ();
    my @pix1Y = ();
    my @pix2X = ();
    my @pix2Y = ();
    foreach (@pixx) {
        my @equ = @$_;

        my (@side, @PP1, @PP2);
        &get_point_chip_side_fixed($_, $limits, \@PP1, \@PP2, \@side_ref);

        push(@pix1X, $PP1[0]);
        push(@pix1Y, $PP1[1]);
        push(@pix2X, $PP2[0]);
        push(@pix2Y, $PP2[1]);

    }

    # Find the extrema
    my (@pp11, @pp12);
    if ($side_ref[0] == 1 || $side_ref[0] == 3) {    # use X
        my @sorted = sort by_number @pix1X;
        if ($side_ref[0] == 1) {
            @pp11 = ($sorted[0], $limits->{"Y_1_1"});
            @pp12 = ($sorted[$#sorted], $limits->{"Y_1_1"});
        } else {
            @pp11 = ($sorted[0], $limits->{"Y_X_Y"});
            @pp12 = ($sorted[$#sorted], $limits->{"Y_X_Y"});
        }
    } else {                                         # use Y
        my @sorted = sort by_number @pix1Y;
        if ($side_ref[0] == 2) {
            @pp11 = ($limits->{"X_X_Y"}, $sorted[0]);
            @pp12 = ($limits->{"X_X_Y"}, $sorted[$#sorted]);
        } else {
            @pp11 = ($limits->{"X_1_1"}, $sorted[0]);
            @pp12 = ($limits->{"X_X_Y"}, $sorted[$#sorted]);
        }
    }

    my (@pp21, @pp22);
    if ($side_ref[1] == 1 || $side_ref[1] == 3) {    # use X
        my @sorted = sort by_number @pix2X;
        if ($side_ref[1] == 1) {
            @pp21 = ($sorted[0], $limits->{"Y_1_1"});
            @pp22 = ($sorted[$#sorted], $limits->{"Y_1_1"});
        } else {
            @pp21 = ($sorted[0], $limits->{"Y_X_Y"});
            @pp22 = ($sorted[$#sorted], $limits->{"Y_X_Y"});
        }
    } else {                                         # use Y
        my @sorted = sort by_number @pix2Y;
        if ($side_ref[1] == 2) {
            @pp21 = ($limits->{"X_X_Y"}, $sorted[0]);
            @pp22 = ($limits->{"X_X_Y"}, $sorted[$#sorted]);
        } else {
            @pp21 = ($limits->{"X_1_1"}, $sorted[0]);
            @pp22 = ($limits->{"X_1_1"}, $sorted[$#sorted]);
        }
    }

    # get the limits + organize the output
    &get_point_order(\@pp11, \@pp12, \@pp21, \@pp22, \@p1, \@p2, \@p3, \@p4);

    $listp->[0]->[0] = $p1[0] + 1;
    $listp->[0]->[1] = $p1[1] + 1;
    $listp->[1]->[0] = $p2[0] + 1;
    $listp->[1]->[1] = $p2[1] + 1;
    $listp->[2]->[0] = $p3[0] + 1;
    $listp->[2]->[1] = $p3[1] + 1;
    $listp->[3]->[0] = $p4[0] + 1;
    $listp->[3]->[1] = $p4[1] + 1;
}

######################################################################
sub add_flag {
    my ($image, $list) = @_;

    my $Nx = `$dfits $image | $fitsort -d NAXIS1 | $awk '{print \$2}'` + 0.0;
    my $Ny = `$dfits $image | $fitsort -d NAXIS2 | $awk '{print \$2}'` + 0.0;

    # Loop on the tracks
    $lpix = "tmppix.listZ";
    open(OUT, ">$lpix");

    foreach (@$list) {
        my $r      = $_->[0];
        my $theta  = $_->[1] - 3.1415927 / 2.0;
        my $thetar = $_->[1];
        my $width  = $_->[2];

	if ($width > 0.0) {
            # Mask lines
            if (abs(cos($theta)) > 0.5) {    # y = ax + b
                $ydec = ($width / 2.0) / cos($theta);
                $a    = -cos($thetar) / sin($thetar);
                $b    = $r / sin($thetar);
                my $ii;
                for ($ii = 1; $ii <= $Nx; $ii++) {
                    $ycurrent = int($a * $ii + $b - 1 - $ydec);
                    if ($ycurrent < 1) {
                        $ycurrent = 1;
                    }
                    $ymax = int($a * $ii + $b + 1 + $ydec);
                    if ($ymax > $Ny) {
                        $ymax = $Ny;
                    }
                    while ($ycurrent <= $ymax) {
                        print OUT "$ii  $ycurrent  1.0 \n";
                        $ycurrent++;
                    }
                }
            } else {    # x = ay + b
                $xdec = ($width / 2.0) / sin($theta);
                $a    = -sin($thetar) / cos($thetar);
                $b    = $r / cos($thetar);
                my $ii;
                for ($ii = 1; $ii <= $Ny; $ii++) {
    
                    $xcurrent = int($a * $ii + $b - 1 - $xdec);
                    if ($xcurrent < 1) {
                        $xcurrent = 1;
                    }
                    $xmax = int($a * $ii + $b + 1 + $xdec);
                    if ($xmax > $Nx) {
                        $xmax = $Nx;
                    }
                    while ($xcurrent <= $xmax) {
                        print OUT "$xcurrent $ii 1.0 \n";
                        $xcurrent++;
                    }
                }
            }
        }
    }
    close OUT;

    # Flag it
    system("$setpix $image \@" . $lpix . "\n");

}
######################################################################
sub get_width {
    my ($image, $lnumber, $eq, $limits) = @_;
    my @ref = (1, 1);

    # get the pixels
    my @pix = ();
    foreach (@$lnumber) {
        my $min = $_ - 0.1;
        $max = $_ + 0.1;
        my @lines = `$getpix -in $image -v $_ \n`;
        foreach (@lines) {
            $_ = /(\d+),(\d+)/;
            push(@pix, [($1 * $drho, ($2 - 1800) * $dtheta)]);
        }
    }

    # which equation of boarder / side ?
    @PP1 = ();
    @PP2 = ();
    my @side;

    @equation = @$eq;

    &get_eqside_chip($eq, $limits, \@PP1, \@PP2, \@side);

    my @pix1X = ();
    my @pix1Y = ();
    my @pix2X = ();
    my @pix2Y = ();

    foreach (@pix) {
        my @equ = @$_;

        # get the intersection points
        my @ref = (1, 1);
        my @P = get_intersect_hough(\@PP1, $_, \@ref);
        push(@pix1X, $P[0]);
        push(@pix1Y, $P[1]);
        my @Q = get_intersect_hough(\@PP2, $_, \@ref);
        push(@pix2X, $Q[0]);
        push(@pix2Y, $Q[1]);
    }

    # get width
    my $theta = $eq->[1];

    my $width1 = -1;
    if ($side[0] == 1 || $side[0] == 3) {    # use X
        my @sorted = sort by_number @pix1X;
        $width1 = ($sorted[$#sorted] - $sorted[0]) * cos($theta);
    } else {                                 # use Y
        my @sorted = sort by_number @pix1Y;
        $width1 = ($sorted[$#sorted] - $sorted[0]) * sin($theta);
    }
    my $width2 = -1;
    if ($side[1] == 1 || $side[1] == 3) {    # use X
        my @sorted = sort by_number @pix2X;
        $width2 = ($sorted[$#sorted] - $sorted[0]) * cos($theta);
    } else {                                 # use Y
        my @sorted = sort by_number @pix2Y;
        $width2 = ($sorted[$#sorted] - $sorted[0]) * sin($theta);
    }

    return ($width1 + $width2) / 2.0;
}
######################################################################
sub get_closest_chips_OLD {
    my ($list, $chip) = @_;

    my %limits = get_chippix_mosaic_instrum($period, $wchip, $instrum);
    $Xref =
      ($limits{'X_1_1'} +
          $limits{'X_X_1'} +
          $limits{'X_1_Y'} +
          $limits{'X_X_Y'}) / 4;
    $Yref =
      ($limits{'Y_1_1'} +
          $limits{'Y_X_1'} +
          $limits{'Y_1_Y'} +
          $limits{'Y_X_Y'}) / 4;

    my @dist  = ();
    my @index = ();
    my %data  = ();
    my $ii    = -1;
    foreach (@$list) {
        my $wchip = $_;
        next if ($wchip == $chip);
        $ii++;

        my %limit = get_chippix_mosaic_instrum($period, $wchip, $instrum);
        $Xcenter =
          ($limit{'X_1_1'} +
              $limit{'X_X_1'} +
              $limit{'X_1_Y'} +
              $limit{'X_X_Y'}) / 4;
        $Ycenter =
          ($limit{'Y_1_1'} +
              $limit{'Y_X_1'} +
              $limit{'Y_1_Y'} +
              $limit{'Y_X_Y'}) / 4;
        $distt =
          ($Xref - $Xcenter) * ($Xref - $Xcenter) +
          ($Yref - $Ycenter) * ($Yref - $Ycenter);

        push(@dist,  $distt);
        push(@index, $ii);
        $data{$ii} = $_;
    }

    @isorted = sort { $data{$a} <=> $data{$b} } @index;

    return ($list->[$isorted[0]], $list->[$isorted[1]]);
}
######################################################################
sub get_closest_chips {
    my ($list, $chip, $moseq) = @_;

    my %limits = get_chippix_mosaic_instrum($period, $chip, $instrum);

    # get the intersection points with the edges on the missing chip
    my (@P1ref, @P2ref);
    &get_point_chip_ref($moseq, \%limits, \@Pref1, \@Pref2, \@ref);
    if ($debug) {
        print "Missing chip : intersect points -> @Pref1 @Pref2 \n";
    }

# compare to the intersection points of surrounding chips and find the closest one
    my %data = ();
    foreach (@$list) {
        my $chipp = $_;
        my %limits = get_chippix_mosaic_instrum($period, $chipp, $instrum);
        my (@P1, @P2);
        &get_point_chip_ref($moseq, \%limits, \@P1, \@P2, \@ref);
        if ($debug) {
            print "Chip $chipp : intersect points -> @P1 @P2 \n";
        }

        $d1 =
          ($P1[0] - $Pref1[0]) * ($P1[0] - $Pref1[0]) +
          ($P1[1] - $Pref1[1]) * ($P1[1] - $Pref1[1]);
        $d2 =
          ($P1[0] - $Pref2[0]) * ($P1[0] - $Pref2[0]) +
          ($P1[1] - $Pref2[1]) * ($P1[1] - $Pref2[1]);
        $d3 =
          ($P2[0] - $Pref2[0]) * ($P2[0] - $Pref2[0]) +
          ($P2[1] - $Pref2[1]) * ($P2[1] - $Pref2[1]);
        $d4 =
          ($P2[0] - $Pref1[0]) * ($P2[0] - $Pref1[0]) +
          ($P2[1] - $Pref1[1]) * ($P2[1] - $Pref1[1]);

        $dmin = min(($d1, $d2, $d3, $d4));

        $data{$_} = $dmin;
        if ($debug) {
            print "dmin $dmin of $d1,$d2,$d3,$d4 \n";
        }
    }

    @csorted = sort { $data{$a} <=> $data{$b} } @$list;

    return ($csorted[0], $csorted[1]);

}
#####################################################################
# find the two closest chips (one for each end of the line)
#####################################################################
sub get_closest_chips_check {
    my ($list, $chip, $moseq) = @_;

    my %limits = get_chippix_mosaic_instrum($period, $chip, $instrum);

    # get the intersection points with the edges on the missing chip
    my (@P1ref, @P2ref);
    &get_point_chip_ref($moseq, \%limits, \@Pref1, \@Pref2, \@ref);

    if ($debug) {
        print "Missing chip : intersect points -> @Pref1 @Pref2 \n";
    }

    # Vertical line ????
    if ($Pref1[0] == $Pref2[0]) {

        # order it
        if ($Pref1[1] > $Pref2[1]) {
            $tmp      = $Pref2[0];
            $Pref2[0] = $Pref1[0];
            $Pref1[0] = $Pref2[0];
            $tmp      = $Pref2[1];
            $Pref2[1] = $Pref1[1];
            $Pref1[1] = $Pref2[1];
        }

        # vertical line
        my @line1 = ();
        my @line2 = ();
        my %data  = ();
        foreach (@$list) {
            my $chipp = $_;
            next if ($chipp == $chip);
            my %limits =
              get_chippix_mosaic_instrum($period, $chipp, $instrum);
            my (@P1, @P2);
            &get_point_chip_ref($moseq, \%limits, \@P1, \@P2, \@ref);
            if ($debug) {
                print "Chip $chipp : intersect points -> @P1 @P2 \n";
            }

            # side1
            if ($P1[1] < $Pref1[1]) {
                $d1 =
                  ($P1[0] - $Pref1[0]) * ($P1[0] - $Pref1[0]) +
                  ($P1[1] - $Pref1[1]) * ($P1[1] - $Pref1[1]);
                $data{$d1} = $chipp;
                push(@line1, $d1);
            }
            if ($P2[1] < $Pref1[1]) {
                $d1 =
                  ($P2[0] - $Pref1[0]) * ($P2[0] - $Pref1[0]) +
                  ($P2[1] - $Pref1[1]) * ($P2[1] - $Pref1[1]);
                $data{$d1} = $chipp;
                push(@line1, $d1);
            }

            # side2
            if ($P1[1] > $Pref2[1]) {
                $d1 =
                  ($P1[0] - $Pref2[0]) * ($P1[0] - $Pref2[0]) +
                  ($P1[1] - $Pref2[1]) * ($P1[1] - $Pref2[1]);
                $data{$d1} = $chipp;
                push(@line2, $d1);
            }
            if ($P2[1] > $Pref2[1]) {
                $d1 =
                  ($P2[0] - $Pref2[0]) * ($P2[0] - $Pref2[0]) +
                  ($P2[1] - $Pref2[1]) * ($P2[1] - $Pref2[1]);
                $data{$d1} = $chipp;
                push(@line2, $d1);
            }

        }

        #side1
        if ($#line1 == -1) {
            $res1 = -1;
        } else {
            @c1sorted = sort { $data{$a} <=> $data{$b} } @line1;
            $res1 = $data{$c1sorted[0]};
        }
        if ($#line2 == -1) {
            $csorted[1] = -1;
        } else {
            @c2sorted = sort { $data{$a} <=> $data{$b} } @line2;
            $res2 = $data{$c2sorted[0]};
        }

        return ($res1, $res2);
    } else {

        # order it
        if ($Pref1[0] > $Pref2[0]) {
            $tmp      = $Pref2[0];
            $Pref2[0] = $Pref1[0];
            $Pref1[0] = $Pref2[0];
            $tmp      = $Pref2[1];
            $Pref2[1] = $Pref1[1];
            $Pref1[1] = $Pref2[1];

        }

        # Non vertical line
        my @line1 = ();
        my @line2 = ();
        my %data  = ();

        foreach (@$list) {
            my $chipp = $_;
            next if ($chipp == $chip);
            my %limits =
              get_chippix_mosaic_instrum($period, $chipp, $instrum);
            my (@P1, @P2);
            &get_point_chip_ref($moseq, \%limits, \@P1, \@P2, \@ref);
            if ($debug) {
                print "Chip $chipp : intersect points -> @P1 @P2 \n";
            }

            # side1
            if ($P1[0] < $Pref1[0]) {
                $d1 =
                  ($P1[0] - $Pref1[0]) * ($P1[0] - $Pref1[0]) +
                  ($P1[1] - $Pref1[1]) * ($P1[1] - $Pref1[1]);
                $data{$d1} = $chipp;
                push(@line1, $d1);
            }
            if ($P2[0] < $Pref1[0]) {
                $d1 =
                  ($P2[0] - $Pref1[0]) * ($P2[0] - $Pref1[0]) +
                  ($P2[1] - $Pref1[1]) * ($P2[1] - $Pref1[1]);
                $data{$d1} = $chipp;
                push(@line1, $d1);
            }

            # side2
            if ($P1[0] > $Pref2[0]) {
                $d1 =
                  ($P1[0] - $Pref2[0]) * ($P1[0] - $Pref2[0]) +
                  ($P1[1] - $Pref2[1]) * ($P1[1] - $Pref2[1]);
                $data{$d1} = $chipp;
                push(@line2, $d1);
            }
            if ($P2[0] > $Pref2[0]) {
                $d1 =
                  ($P2[0] - $Pref2[0]) * ($P2[0] - $Pref2[0]) +
                  ($P2[1] - $Pref2[1]) * ($P2[1] - $Pref2[1]);
                $data{$d1} = $chipp;
                push(@line2, $d1);
            }
        }

        #side1
        if ($#line1 == -1) {
            $res1 = -1;
        } else {
            @c1sorted = sort by_number @line1;
            $res1     = $data{$c1sorted[0]};
        }
        if ($#line2 == -1) {
            $csorted[1] = -1;
        } else {
            @c2sorted = sort by_number @line2;
            $res2     = $data{$c2sorted[0]};
        }

        return ($res1, $res2);
    }
}

#####################################################################
sub by_number {
    if ($a < $b) {
        return -1;
    } elsif ($a == $b) {
        return 0;
    } elsif ($a > $b) {
        return 1;
    }
}

#####################################################################
sub new_center {
    my ($lgr, $sat_moseq, $ig) = @_;

    my @res  = ();
    my @res2 = ();

    my $n  = 0;
    my $gr = $lgr->[$ig];
    foreach (@{$gr->{'L'}}) {
        $n++;
        my @eq = @{$sat_moseq->[$_]};
        $res[0] += $eq[0];
        $res[1] += $eq[1];

        $res2[0] += $eq[0] * $eq[0];
        $res2[1] += $eq[1] * $eq[1];
    }

    $res[0] /= $n;
    $res[1] /= $n;

    $res2[0] /= $n;
    $res2[1] /= $n;
    $res2[0] = sqrt($res2[0] - $res[0] * $res[0]);
    $res2[1] = sqrt($res2[1] - $res[1] * $res[1]);

    $gr->{'C'} = [@res];
    $gr->{'D'} = [@res2];
}

#####################################################################
sub gdist {
    my ($eq, $groups, $ig) = @_;

    my @geq = @{$groups->[$ig]->{'C'}};

    $dist =
      ($eq->[0] - $geq[0]) *
      ($eq->[0] - $geq[0]) +
      ($eq->[1] / $dtheta - $geq[1] / $dtheta) *
      ($eq->[1] / $dtheta - $geq[1] / $dtheta);

    return $dist;
}
#####################################################################
# Check is similar strack equation are on the mosaic (other chips)
#####################################################################
sub check_global_pix {
    my ($eq, $leq, $lchip, $rad, $rchip, $im, $num, $ref) = @_;

    my %limits_chip = get_chip_limits_instrum_ind0($period, $rchip, $instrum);
    my %limits_mos = get_chippix_mosaic_instrum($period, $rchip, $instrum);

    # get the pixels of this track
    my @list = `$getpix -in $im -v $num  \n`;
    my @data = ();
    foreach (@list) {
        my $line = $_;
        if ($line =~ /(\d+),(\d+).*=\s(\S+)/) {

            my $eq1 = ($1 - 1) * $drho;
            my $eq2 = ($2 - 1 - $Ny / 2 + 0.5) * $dtheta;

            push(@data, [($eq1, $eq2)]);
        }
    }

    # Convert px to moseq + check
    foreach (@data) {
        my @tt = @$_;

        # convert
        my @eq = ();
        &trans_line(\%limits_chip, \%limits_mos, \@tt, \@eq, $ref);

        my @rref = @$ref;

        my $test = check_global(\@eq, $leq, $lchip, $rad, $rchip);
        if ($test == 0) {
            return 0;
        }
    }

    return 1;

}
#####################################################################
# Check is similar strack equation are on the mosaic (other chips)
#####################################################################
sub check_global {
    my ($eq, $leq, $lchip, $rad, $refchip) = @_;
    my $rad2   = $rad * $rad;
    my $dtheta = 3.1415927 / 1800;

    my @list_chip = ();
    my $ii        = -1;
    foreach (@$leq) {
        $ii++;
        my $chip = $lchip->[$ii];
        if (!is_in_num(\@list_chip, $chip)) {
            push(@list_chip, $chip);
        }
    }
    my $nchip = scalar(@list_chip);
    if ($nchip >= 4) {
        $ndetect = 2;
    } else {
        $ndetect = 1;
    }

    my @dchip = ();
    my $ii    = -1;
    foreach (@$leq) {

        my @tmp = @$eq;

        $ii++;
        my $chip = $lchip->[$ii];
        $dist =
          ($_->[0] - $eq->[0]) *
          ($_->[0] - $eq->[0]) +
          ($_->[1] / $dtheta - $eq->[1] / $dtheta) *
          ($_->[1] / $dtheta - $eq->[1] / $dtheta);

        if ($chip != $refchip && $dist > 0 && $dist < $rad2) {
            if (!is_in_num(\@dchip, $chip)) {
                push(@dchip, $chip);
            }
            if (scalar(@dchip) >= $ndetect) {
                return 0;
            }
        }
    }
    return 1;
}

######################################################################
sub filter_detect {
    my ($data, $thresh, $l1, $l2) = @_;

    $ii = -1;
    foreach (@{$data->{'NUMBER'}}) {
        $ii++;
        if ($data->{'FLUX_MAX'}->[$ii] >= $thresh) {
            push(@$l2, $ii);
        } else {
            push(@$l1, $ii);
        }
    }
}

#######################################################################
sub write_param {
    open(OUT, ">default.param_Z_$$");
    print OUT <<END;
   NUMBER
   XMIN_IMAGE
   YMIN_IMAGE
   XMAX_IMAGE
   YMAX_IMAGE
   XPEAK_IMAGE
   YPEAK_IMAGE
   FLUX_MAX
   ISOAREA_IMAGE
   X_IMAGE
   Y_IMAGE
   ELLIPTICITY
   FLAGS
   CLASS_STAR
END
    close OUT;

    open(OUT, ">default.conv_Z_$$");
    print OUT <<END;
CONV NORM
# 3x3 ``all-ground'' convolution mask with FWHM = 2 pixels.
1 2 1
2 4 2
1 2 1
END
    close OUT;

    open(OUT, ">default.nnw_Z_$$");
    print OUT <<END;
NNW  
# Neural Network Weights for the SExtractor star/galaxy classifier (V1.3)
# inputs:       9 for profile parameters + 1 for seeing.
# outputs:      ``Stellarity index'' (0.0 to 1.0)
# Seeing FWHM range: from 0.025 to 5.5'' (images must have 1.5 < FWHM < 5 pixels)
# Optimized for Moffat profiles with 2<= beta <= 4.

 3 10 10  1

-1.56604e+00 -2.48265e+00 -1.44564e+00 -1.24675e+00 -9.44913e-01 -5.22453e-01  4.61342e-02  8.31957e-01  2.15505e+00  2.64769e-01
 3.03477e+00  2.69561e+00  3.16188e+00  3.34497e+00  3.51885e+00  3.65570e+00  3.74856e+00  3.84541e+00  4.22811e+00  3.27734e+00

-3.22480e-01 -2.12804e+00  6.50750e-01 -1.11242e+00 -1.40683e+00 -1.55944e+00 -1.84558e+00 -1.18946e-01  5.52395e-01 -4.36564e-01 -5.30052e+00
 4.62594e-01 -3.29127e+00  1.10950e+00 -6.01857e-01  1.29492e-01  1.42290e+00  2.90741e+00  2.44058e+00 -9.19118e-01  8.42851e-01 -4.69824e+00
-2.57424e+00  8.96469e-01  8.34775e-01  2.18845e+00  2.46526e+00  8.60878e-02 -6.88080e-01 -1.33623e-02  9.30403e-02  1.64942e+00 -1.01231e+00
 4.81041e+00  1.53747e+00 -1.12216e+00 -3.16008e+00 -1.67404e+00 -1.75767e+00 -1.29310e+00  5.59549e-01  8.08468e-01 -1.01592e-02 -7.54052e+00
 1.01933e+01 -2.09484e+01 -1.07426e+00  9.87912e-01  6.05210e-01 -6.04535e-02 -5.87826e-01 -7.94117e-01 -4.89190e-01 -8.12710e-02 -2.07067e+01
-5.31793e+00  7.94240e+00 -4.64165e+00 -4.37436e+00 -1.55417e+00  7.54368e-01  1.09608e+00  1.45967e+00  1.62946e+00 -1.01301e+00  1.13514e-01
 2.20336e-01  1.70056e+00 -5.20105e-01 -4.28330e-01  1.57258e-03 -3.36502e-01 -8.18568e-02 -7.16163e+00  8.23195e+00 -1.71561e-02 -1.13749e+01
 3.75075e+00  7.25399e+00 -1.75325e+00 -2.68814e+00 -3.71128e+00 -4.62933e+00 -2.13747e+00 -1.89186e-01  1.29122e+00 -7.49380e-01  6.71712e-01
-8.41923e-01  4.64997e+00  5.65808e-01 -3.08277e-01 -1.01687e+00  1.73127e-01 -8.92130e-01  1.89044e+00 -2.75543e-01 -7.72828e-01  5.36745e-01
-3.65598e+00  7.56997e+00 -3.76373e+00 -1.74542e+00 -1.37540e-01 -5.55400e-01 -1.59195e-01  1.27910e-01  1.91906e+00  1.42119e+00 -4.35502e+00

-1.70059e+00 -3.65695e+00  1.22367e+00 -5.74367e-01 -3.29571e+00  2.46316e+00  5.22353e+00  2.42038e+00  1.22919e+00 -9.22250e-01 -2.32028e+00

 0.00000e+00
 1.00000e+00
END
    close OUT;

    open(OUT, "> default.sex_Z_$$");
    print OUT <<END;

# Default configuration file for SExtractor V1.2b14 - > 2.0
#
# (*) indicates parameters which can be omitted from this config file.

#-------------------------------- Catalog ------------------------------------

CATALOG_NAME    outfile        # name of the output catalog
CATALOG_TYPE    ASCII_HEAD      # "NONE","ASCII_HEAD","ASCII","FITS_1.0"
                                # or "FITS_LDAC"

PARAMETERS_NAME default.param_Z_$$     # name of the file containing catalog contents

#------------------------------- Extraction ----------------------------------

DETECT_TYPE     CCD             # "CCD" or "PHOTO" (*)
FLAG_IMAGE      flag.fits       # filename for an input FLAG-image
DETECT_MINAREA  1               # minimum number of pixels above threshold
THRESH_TYPE     ABSOLUTE
DETECT_THRESH   $mthresh          # <sigmas> or <threshold>,<ZP> in mag.arcsec-2
ANALYSIS_THRESH $mthresh             # <sigmas> or <threshold>,<ZP> in mag.arcsec-2

FILTER          Y               # apply filter for detection ("Y" or "N")?
FILTER_NAME     default.conv_Z_$$            # name of the file containing the filter

DEBLEND_NTHRESH 32              # Number of deblending sub-thresholds
DEBLEND_MINCONT 0.005           # Minimum contrast parameter for deblending

CLEAN           Y               # Clean spurious detections? (Y or N)?
CLEAN_PARAM     1.0             # Cleaning efficiency

MASK_TYPE       CORRECT         # type of detection MASKing: can be one of
                                # "NONE", "BLANK" or "CORRECT"

#------------------------------ Photometry -----------------------------------

PHOT_APERTURES  30               # MAG_APER aperture diameter(s) in pixels
PHOT_AUTOPARAMS 2.5, 3.5        # MAG_AUTO parameters: <Kron_fact>,<min_radius>

SATUR_LEVEL     50000          # level (in ADUs) at which arises saturation

MAG_ZEROPOINT   30.186630274887          # magnitude zero-point
MAG_GAMMA       4.0             # gamma of emulsion (for photographic scans)
GAIN            7.0             # detector gain in e-/ADU.
PIXEL_SCALE     0.1               # size of pixel in arcsec (0=use FITS WCS info).

#------------------------- Star/Galaxy Separation ----------------------------

SEEING_FWHM     1.0           # stellar FWHM in arcsec
STARNNW_NAME    default.nnw_Z_$$            # Neural-Network_Weight table filename

#------------------------------ Background -----------------------------------

BACK_SIZE       64              # Background mesh: <size> or <width>,<height>
BACK_FILTERSIZE 4               # Background filter: <size> or <width>,<height>

BACKPHOTO_TYPE  GLOBAL          # can be "GLOBAL" or "LOCAL" (*)
BACKPHOTO_THICK 24              # thickness of the background LOCAL annulus(*)

#------------------------------ Check Image ----------------------------------

CHECKIMAGE_TYPE SEGMENTATION    # can be one of "NONE", "BACKGROUND",
                                # "MINIBACKGROUND", "-BACKGROUND", "OBJECTS",
                                # "-OBJECTS", "SEGMENTATION", "APERTURES",
                                # or "FILTERED" (*)
CHECKIMAGE_NAME check.fits      # Filename for the check-image (*)

#--------------------- Memory (change with caution!) -------------------------

MEMORY_OBJSTACK 4000            # number of objects in stack
MEMORY_PIXSTACK 4000000          # number of pixels in stack
MEMORY_BUFSIZE  1024            # number of lines in buffer

#----------------------------- Miscellaneous ---------------------------------

VERBOSE_TYPE    NORMAL          # can be "QUIET", "NORMAL" or "FULL" (*)

#------------------------------- New Stuff -----------------------------------
WEIGHT_TYPE  NONE
END
    close OUT;

}

#####################################################################
sub filter_close_sub {
    my ($group, $data, $tol, $list_xy) = @_;

    my $ngroups = scalar(@$group);

    my $tol2 = $tol * $tol;

   # Check if some elements are in several groups --> if yes, unify the groups
    my @groups_list = ();
    my %unify       = ();

    # Loop on the groups
    $ig = -1;
    foreach (@$group) {
        $ig++;

        # Loop on the elements
        $flag = 0;
        foreach (@{$_->{"L"}}) {
            $ind = $_;

            # Check if present in the %groups_list
            $tg = -1;
            foreach (@groups_list) {
                $tg++;
                $index = $_;

                # Check if already present
                if (is_in_num($_, $ind)) {
                    $flag = 1;
                    last;
                }

                # Check minimum distance
                $xref = $data->{"X_IMAGE"}->[$ind];
                $yref = $data->{"Y_IMAGE"}->[$ind];
                foreach (@$_) {
                    $xt = $data->{"X_IMAGE"}->[$_];
                    $yt = $data->{"Y_IMAGE"}->[$_];

                    if (($xt - $xref) * ($xt - $xref) +
                        ($yt - $yref) * ($yt - $yref) <= $tol2)
                    {
                        $flag = 1;
                        print "    FOUND \n";
                        last;
                    }
                }

                if ($flag == 1) {
                    last;
                }

            }

            # ALREADY EXISTING GROUP
            if ($flag == 1) {
                $unify{$ig} = $tg;
                foreach (@{$group->[$ig]->{"L"}}) {
                    push(@{$groups_list[$tg]}, $_);
                }
                last;
            }
        }

        # New group
        if ($flag == 0) {
            $n = scalar(@groups_list);
            $groups_list[$n] = [@{$group->[$ig]->{"L"}}];
        }
    }

    if (defined $debug) {
        print "\n\n RECHECK ... \n";
        foreach (@groups_list) {
            print "New_group : ";
            foreach (@$_) {
                print " $_ ";
            }
            print "\n";
        }
        print "\n\n UNIFY ... \n";
        @keys = keys(%unify);
        foreach (@keys) {
            print "unify $_ in " . $unify{$_} . "\n";
        }
        print "\n\n";
    }

    # rebuild the groups
    @$group = ();
    $ig     = -1;
    foreach (@groups_list) {
        $ig++;
        $group->[$ig]->{"L"} = [@$_];
        &new_center($group, $list_xy, $ig);
    }

    my $ngroups2 = scalar(@$group);
    if ($ngroups2 != $ngroups) {
        return 2;
    } else {
        return 0;
    }

}

#####################################################################
sub filter_close {
    my ($im, $data, $tol, $group, $list_xy) = @_;

    my $tol2 = $tol * $tol;

    my $ii = -1;

    foreach (@{$data->{"NUMBER"}}) {
        $ii++;
        $xpeak = $data->{"X_IMAGE"}->[$ii];
        $ypeak = $data->{"Y_IMAGE"}->[$ii];
        push(@$list_xy, [($xpeak, $ypeak)]);
    }

    #my @group;
    $ii = -1;
    my $ng = 0;
    foreach (@{$data->{"NUMBER"}}) {
        $ii++;
        $segnum = $_;
        $xpeak  = $data->{"X_IMAGE"}->[$ii];
        $ypeak  = $data->{"Y_IMAGE"}->[$ii];

        $flag = 0;
        $ig   = -1;
        foreach (@$group) {
            $gr = $_;
            $ig++;
            $xm = $gr->{"C"}->[0];
            $ym = $gr->{"C"}->[1];

            my $dist =
              ($xpeak - $xm) * ($xpeak - $xm) +
              ($ypeak - $ym) * ($ypeak - $ym);

            # New element
            if ($dist <= $tol2) {
                push(@{$gr->{"L"}}, $ii);
                &new_center($group, $list_xy, $ig);
                $flag = 1;
            }
        }

        # new group
        if ($flag == 0) {
            $group->[$ng]->{"C"} = [($xpeak, $ypeak)];
            push(@{$group->[$ng]->{"L"}}, $ii);
            $ng++;
        }
    }

    my $ff = 1;
    while ($ff > 0) {
        $ff = filter_close_sub($group, $data, $tol, $list_xy);
    }

    # find the object with maximum flux in each group
    if ($debug) {
        print "GROUPS : \n";
    }

    my $flag = 0;
    foreach (@$group) {
        $gr   = $_;
        $max  = -1;
        $imax = -1;
        foreach (@{$gr->{"L"}}) {
            if ($data->{"FLUX_MAX"}->[$_] > $max) {
                $max  = $data->{"FLUX_MAX"}->[$_];
                $imax = $_;
            }
            if ($debug) {
                print " $_ ";
            }
        }
        if ($debug) {
            print "\n";
        }

        $gr->{"IPEAK"}   = $imax;
        $gr->{"NUMPEAK"} = $data->{"NUMBER"}->[$imax];

        if (scalar(@{$gr->{"L"}}) > 1) {
            $flag = 1;
        }
    }

    return $flag;
}

####################################################################
# Find member of each group with max flux
####################################################################
sub find_maxflux {
    my ($groups, $data) = @_;

    my $flag = 0;
    foreach (@$groups) {
        my $gr   = $_;
        my $max  = -1;
        my $imax = -1;
        foreach (@{$gr->{"L"}}) {
            if ($data->{"FLUX_MAX"}->[$_] > $max) {
                $max  = $data->{"FLUX_MAX"}->[$_];
                $imax = $_;
            }

        }

        $gr->{"FMAX"}    = $max;
        $gr->{"IPEAK"}   = $imax;
        $gr->{"NUMPEAK"} = $data->{"NUMBER"}->[$imax];

    }
}

#####################################################################
# Filter the seg image according to groups
#####################################################################
sub filter_seg {
    my ($group, $im, $data_sex) = @_;

    # Loop on the groups
    foreach (@$group) {
        $gr = $_;
        if (scalar(@{$gr->{"L"}}) == 1) { next; }

        $num  = $gr->{"NUMPEAK"};
        $minx = 1e6;
        $maxx = -1e6;
        $miny = 1e6;
        $maxy = -1e6;

        my @list_ind = ();
        foreach (@{$gr->{"L"}}) {
            if ($minx > $data_sex->{"XMIN_IMAGE"}->[$_]) {
                $minx = $data_sex->{"XMIN_IMAGE"}->[$_];
            }
            if ($miny > $data_sex->{"YMIN_IMAGE"}->[$_]) {
                $miny = $data_sex->{"YMIN_IMAGE"}->[$_];
            }
            if ($maxx < $data_sex->{"XMAX_IMAGE"}->[$_]) {
                $maxx = $data_sex->{"XMAX_IMAGE"}->[$_];
            }
            if ($maxy < $data_sex->{"YMAX_IMAGE"}->[$_]) {
                $maxy = $data_sex->{"YMAX_IMAGE"}->[$_];
            }
            push(@list_ind, $data_sex->{"NUMBER"}->[$_]);
        }

        # replace on the image
        $minx--;
        $miny--;
        $maxx++;
        $maxy++;
        @pix = `$getpix -in $im -vmin 0.5 \n`;

        $flagf = "tmp_flag.ZZ_$$";

        if (-e $flagf) {
            unlink($flagf);
        }

        open(OUT, ">" . $flagf);
        foreach (@pix) {
            $line = $_;

            if ($line =~ /\[(\d+),(\d+)\]\s+=\s+(\S+)/) {
                if (is_in_num(\@list_ind, $3)) {
                    print OUT "$1 $2 $num \n";
                }
            } else {
                die "Problem with line $line \n";
            }
        }
        close OUT;

        # Do it
        system("$setpix $im \@" . $flagf . "  \n");

        # replace in data_sex
        $data_sex->{"XMIN_IMAGE"}->[$num] = $minx;
        $data_sex->{"XMAX_IMAGE"}->[$num] = $maxx;
        $data_sex->{"YMIN_IMAGE"}->[$num] = $miny;
        $data_sex->{"YMAX_IMAGE"}->[$num] = $maxy;
    }
}

#####################################################################
# Filter the indexes according to groups (put only the groups in the list)
#####################################################################
sub filter_all {
    my ($all, $group) = @_;

    my @all1 = @$all;
    @$all = ();

    my @list_ind = ();
    foreach (@$group) {
        my $gr = $_;
        if (scalar(@{$gr->{"L"}}) == 1) { next; }

        $imax = $gr->{"IPEAK"};
        foreach (@{$gr->{"L"}}) {
            if ($_ != $imax) {
                push(@list_ind, $_);
            }
        }
    }

    foreach (@all1) {
        if (!is_in_num(\@list_ind, $_)) {
            push(@$all, $_);
        }
    }
}

#####################################################################
# Check the badccds in a list
#####################################################################
sub check_badccd {
    my ($list, $lbad) = @_;

    my @list2;
    @list2 = @$list;

    @$list = ();
    foreach (@list2) {
        if (!is_in_num($lbad, $_)) {
            push(@$list, $_);
        }
    }
}

#####################################################################
# Min size of the track
#####################################################################
sub min_size {
    my ($listp, $side, $min_width) = @_;
    my @list = @$listp;

    # Top bottom sides
    if ($side->[0] == 1) {

        # find the top/bottom points
        my @top    = ();
        my @bottom = ();
        my $il     = -1;
        foreach (@$listp) {
            $il++;
            if ($_->[1] < 2) {
                push(@bottom, $il);
            } else {
                push(@top, $il);
            }
        }

        if (scalar(@top) != 2 || scalar(@bottom) != 2) {
            print "Big Problem \n";
            die;
        }

        # center top/bottom
        my $top0    = $top[0];
        my $top1    = $top[1];
        my $bottom0 = $bottom[0];
        my $bottom1 = $bottom[1];

        my $xtop        = ($list[$top0]->[0] + $list[$top1]->[0]) / 2.0;
        my $xbottom     = ($list[$bottom0]->[0] + $list[$bottom1]->[0]) / 2.0;
        my $ytop        = $list[$top0]->[1];
        my $ybottom     = $list[$bottom0]->[1];
        my $angle       = atan2($ytop - $ybottom, abs($xtop - $xbottom));
        my $angle_coeff = sin($angle);

        # Check the width of top
        my $widtht =
          abs($list[$top0]->[0] - $list[$top1]->[0]) * $angle_coeff;
        my $widthb =
          abs($list[$bottom0]->[0] - $list[$bottom1]->[0]) * $angle_coeff;
        if ($widtht < $min_width) {
            if ($list[$top0]->[0] < $xtop) {
                $listp->[$top0]->[0] = $xtop - $min_width / $angle_coeff;
                $listp->[$top1]->[0] = $xtop + $min_width / $angle_coeff;
            } else {
                $listp->[$top0]->[0] = $xtop + $min_width / $angle_coeff;
                $listp->[$top1]->[0] = $xtop - $min_width / $angle_coeff;
            }
        }
        if ($widthb < $min_width) {
            if ($list[$bottom0]->[0] < $xbottom) {
                $listp->[$bottom0]->[0] =
                  $xbottom - $min_width / $angle_coeff / 2;
                $listp->[$bottom1]->[0] =
                  $xbottom + $min_width / $angle_coeff / 2;
            } else {
                $listp->[$bottom0]->[0] =
                  $xbottom + $min_width / $angle_coeff / 2;
                $listp->[$bottom1]->[0] =
                  $xbottom - $min_width / $angle_coeff / 2;
            }
        }

        # Right Left sides
    } else {

        # find the top/bottom points
        my @right = ();
        my @left  = ();
        my $il    = -1;
        foreach (@$listp) {
            $il++;
            if ($_->[0] < 2) {
                push(@left, $il);
            } else {
                push(@right, $il);
            }
        }

        if (scalar(@right) != 2 || scalar(@left) != 2) {
            print "Big Problem \n";
            die;
        }

        # center top/bottom
        my $right0 = $right[0];
        my $right1 = $right[1];
        my $left0  = $left[0];
        my $left1  = $left[1];

        my $yright = ($list[$right0]->[1] + $list[$right1]->[1]) / 2.0;
        my $yleft  = ($list[$left0]->[1] + $list[$left1]->[1]) / 2.0;
        my $xright = $list[$right0]->[0];
        my $xleft  = $list[$left0]->[0];

        my $angle = atan2(abs($yleft - $yright), $xright - $xleft);
        my $angle_coeff = cos($angle);

        # Check the width of top
        my $widthl =
          abs($list[$left0]->[1] - $list[$left1]->[1]) * $angle_coeff;
        my $widthr =
          abs($list[$right0]->[1] - $list[$right1]->[1]) * $angle_coeff;
        if ($widthl < $min_width) {
            if ($list[$left0]->[1] < $yleft) {
                $list[$left0]->[1] = $yleft - $min_width / $angle_coeff / 2;
                $list[$left1]->[1] = $yleft + $min_width / $angle_coeff / 2;
            } else {
                $list[$left0]->[1] = $yleft + $min_width / $angle_coeff / 2;
                $list[$left1]->[1] = $yleft - $min_width / $angle_coeff / 2;
            }
        }
        if ($widthr < $min_width) {
            if ($list[$right0]->[1] < $yright) {
                $list[$right0]->[1] = $yright - $min_width / $angle_coeff / 2;
                $list[$right1]->[1] = $yright + $min_width / $angle_coeff / 2;
            } else {
                $list[$right0]->[1] = $yright + $min_width / $angle_coeff / 2;
                $list[$right1]->[1] = $yright - $min_width / $angle_coeff / 2;
            }
        }
    }
}

################################################################################################
sub check_vertical_layout {
    my ($group, $layout, $sat_flag, $sat_chip) = @_;

    my @chips = ();
    foreach (@{$group->{"L"}}) {
        push(@chips, $sat_chip[$_]);
    }

    my $flag = 1;
    my @chiplist;
    foreach (@{$group->{"L"}}) {
        $testchip = $sat_chip[$_];
        $vflag = $sat_flag[$_];
        @chiplist = get_layout_pos($sat_chip[$_], $layout);

        $ncoords = 0 + @chiplist / 2;

        if ($vflag == 1 && $flag == 1) { # Vertical check the other chips
            for ($coord = 0; $coord <= $ncoords; $coord++) {
                $xpos = @chiplist[2 * $coord];
                $ypos = @chiplist[2 * $coord + 1];
                if (exists($layout->{$xpos}->{$ypos - 1})) {
                    $compchip = "$layout->{$xpos}->{$ypos - 1}";
                    $compchip =~ s/_.*//;
                    $compchip_num = $compchip + 0;
                    if (is_in_num(\@chips, $compchip_num) && 
                        $compchip_num != $testchip) {
                        $flag = 0;
                        last;
                    }
                }
                if (exists($layout->{$xpos}->{$ypos + 1})) {
                    $compchip = "$layout->{$xpos}->{$ypos + 1}";
                    $compchip =~ s/_.*//;
                    $compchip_num = $compchip + 0;
                    if (is_in_num(\@chips, $compchip_num) &&
                        $compchip_num != $testchip) {
                        $flag = 0;
                        last;
                    }
                }
            }

        } else {    # Horizontal
            for ($coord = 0; $coord <= $ncoords; $coord++) {
                $xpos = @chiplist[2 * $coord];
                $ypos = @chiplist[2 * $coord + 1];

                if (exists($layout->{$xpos - 1}->{$ypos})) {
                    $compchip = "$layout->{$xpos - 1}->{$ypos}";
                    $compchip =~ s/_.*//;
                    $compchip_num = $compchip + 0;
                    if (is_in_num(\@chips, $compchip_num) &&
                        $compchip_num != $testchip) {
                        $flag = 0;
                        last;
                    }
                }
                if (exists($layout->{$xpos + 1}->{$ypos})) {
                    $compchip = "$layout->{$xpos + 1}->{$ypos}";
                    $compchip =~ s/_.*//;
                    $compchip_num = $compchip + 0;
                    if (is_in_num(\@chips, $compchip_num) &&
                        $compchip_num != $testchip) {
                        $flag = 0;
                        last;
                    }
                }
            }
        }
    }

    return $flag;
}

################################################################################################

